#include <Arduino.h>

// Value Helpers
int16_t make16(uint8_t u1, uint8_t u2)
{
    return ((int16_t)u1 << 8) | u2;
}

uint8_t split16(uint16_t value, uint8_t index)
{
    switch (index)
    {
    case 0:
        return (value >> 8);
    case 1:
        return value & 0xff;
    }
    return 0;
}
#define CLAMP(x, low, high) (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))

#define DEFAULT_SPEED_COEFFICIENT 16384 // Default for SPEED_COEFFICIENT 1.0f [-] higher value == stronger. [0, 65535] = [-2.0 - 2.0]. In this case 16384 = 1.0 * 2^14
#define DEFAULT_STEER_COEFFICIENT 8192  // Defualt for STEER_COEFFICIENT 0.5f [-] higher value == stronger. [0, 65535] = [-2.0 - 2.0]. In this case  8192 = 0.5 * 2^14. If you do not want any steering, set it to 0.

/* mixerFcn(rtu_speed, rtu_steer, &rty_speedR, &rty_speedL);
 * Inputs:       rtu_speed, rtu_steer                  = fixdt(1,16,4)
 * Outputs:      rty_speedR, rty_speedL                = int16_t
 * Parameters:   SPEED_COEFFICIENT, STEER_COEFFICIENT  = fixdt(0,16,14)
 */
void mixerFcn(int16_t rtu_speed, int16_t rtu_steer, int16_t *rty_speedR, int16_t *rty_speedL)
{
    int16_t prodSpeed;
    int16_t prodSteer;
    int32_t tmp;

    prodSpeed = (int16_t)((rtu_speed * (int16_t)DEFAULT_SPEED_COEFFICIENT) >> 14);
    prodSteer = (int16_t)((rtu_steer * (int16_t)DEFAULT_STEER_COEFFICIENT) >> 14);

    tmp = prodSpeed - prodSteer;
    tmp = CLAMP(tmp, -32768, 32767);               // Overflow protection
    *rty_speedR = (int16_t)(tmp >> 4);             // Convert from fixed-point to int
    *rty_speedR = CLAMP(*rty_speedR, -1000, 1000); // TODO: Use constants here

    tmp = prodSpeed + prodSteer;
    tmp = CLAMP(tmp, -32768, 32767);               // Overflow protection
    *rty_speedL = (int16_t)(tmp >> 4);             // Convert from fixed-point to int
    *rty_speedL = CLAMP(*rty_speedL, -1000, 1000); // TODO: Use constants here
}
